@@ -895,8 +895,8 @@ func (f GGML) FlashAttention() bool {
 	return slices.Contains([]string{
 		"gemma3",
 		"gptoss", "gpt-oss",
-		"qwen3",
-		"qwen3moe",
+		"qwen3", "qwen3moe",
+		"qwen3vl", "qwen3vlmoe",
 	}, f.KV().String("general.architecture"))
 }
 
